import sasktranif.sasktranif as skif
import numpy as np
from sasktran.handles import standard_handles
from sasktran.exceptions import wrap_skif_functionfail, SasktranError
from abc import ABCMeta, abstractmethod
from sasktran.util import ArrayWithCallback
import logging
from sasktran import config


class ClimatologyBase(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def skif_object(self, **kwargs) -> skif.ISKClimatology:
        pass


class Climatology(ClimatologyBase):
    """
    A climatology quantifying something in/about the atmosphere. For example climatologies could be used
    to define profiles for temperature, pressure, gas number densities, etc.

    Parameters
    ----------
    name : str
        Name of the climatology to create

    Examples
    --------
    >>> import sasktran as sk
    >>> msis = sk.Climatology('msis90')
    >>> print(msis)
    SasktranIF Climatology: msis90
    Supported Species: ['SKCLIMATOLOGY_PRESSURE_PA', 'SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', 'SKCLIMATOLOGY_TEMPERATURE_K', 'SKCLIMATOLOGY_O2_CM3']
    """
    @wrap_skif_functionfail
    def __init__(self, name: str):
        self._iskclimatology = skif.ISKClimatology(name)
        self._name = name

    def __repr__(self):
        supported_species = self.supported_species()

        representation = ('SasktranIF Climatology: {}\n'
                          'Supported Species: {}'.format(self._name, supported_species))

        return representation

    def skif_object(self, **kwargs) -> skif.ISKClimatology:
        """
        
        Returns
        -------
        skif.ISKClimatology
            Underlying ISKClimatology object which represents the climatology
        """
        return self._iskclimatology

    def supported_species(self):
        """
        Get a list of :ref:`species` identifiers that are supported by this climatology.
         
        Returns
        -------
        list
            Species identifiers that are supported by this climatology. Identifiers correspond to 
            :py:attr:`sasktran.Species.species` 
        """
        supported_species = []

        for handle in standard_handles():
            try:
                self.get_parameter(handle, 0, 0, 1000, 54372)
                supported_species.append(handle)
            except SasktranError:
                pass

        return supported_species

    @wrap_skif_functionfail
    def get_parameter(self, species: str, latitude: float, longitude: float, altitudes: np.ndarray, mjd: float):
        """
        Get a profile from the climatology.

        Parameters
        ----------
        species : str
            Species identifier of the species to query. Corresponds to :py:attr:`sasktran.Species.species`.
        latitude : float
            Latitude in degrees.
        longitude : float
            Longitude in degrees.
        altitudes : np.ndarray
            Profile altitude grid in meters.
        mjd : float
            Modified julian date.
        
        Returns
        -------
        np.ndarray
            Climatology values at the specified altitudes.
        """
        location = [latitude, longitude, 0, mjd]
        self._iskclimatology.UpdateCache(location)
        geo_instant = skif.GEODETIC_INSTANT()
        geo_instant.FromSequence(location)
        return self._iskclimatology.GetHeightProfile(species, geo_instant, np.asarray(altitudes))[1]


class ClimatologyUserDefined(Climatology):
    """
    A special climatology which handles user defined values.

    Parameters
    ----------
    altitudes: np.ndarray
        Array of altitudes in meters that the profile is to be specified at.  Same shape as values.
    values: dict
        dictionary of species, value pairs. Values should be the same length as altitudes.
    interp: str, optional
        One of 'linear' or 'log'. Defines the interpolation space. Default is 'linear'
    spline: bool, optional
        One of True of False.  If True then a quadratic spline will be used to interpolate, if False then piecewise
        linear interpolation is done.  Default is False.

    Examples
    --------
    >>> from sasktran.climatology import ClimatologyUserDefined
    >>> import numpy as np
    >>> altitudes = np.linspace(500, 99500, 100)
    >>> values = np.ones_like(altitudes)
    >>> climatology = ClimatologyUserDefined(altitudes, {'SKCLIMATOLOGY_O3_CM3': values})
    >>> climatology.get_parameter('SKCLIMATOLOGY_O3_CM3', latitude=0, longitude=0, altitudes=[10000, 11000], mjd=54372)
    array([ 1.,  1.])
    >>> climatology['SKCLIMATOLOGY_O3_CM3'] *= 2
    >>> climatology.get_parameter('SKCLIMATOLOGY_O3_CM3', latitude=0, longitude=0, altitudes=[10000, 11000], mjd=54372)
    array([ 2.,  2.])
    """
    def __init__(self, altitudes: np.ndarray, values: dict, interp: str='linear', spline: bool=False):
        super().__init__('USERDEFINED_PROFILE')

        self._altitudes = altitudes
        self._values = {key: ArrayWithCallback(val) for key, val in values.items()}

        for val in self._values.values():
            val.add_modified_callback(self._update_internal_clim)

        self._interp = interp
        self._spline = spline

        self._climatology_created = False
        self._create_climatology()

    def supported_species(self):
        return list(self._values.keys())

    @wrap_skif_functionfail
    def _create_climatology(self):
        if self._interp.lower() == 'linear':
            self._iskclimatology.SetProperty('DoLogInterpolation', 0)
        elif self._interp.lower() == 'log':
            self._iskclimatology.SetProperty('DoLogInterpolation', 1)
        else:
            raise ValueError('interp must be one of linear or log')

        if self._spline:
            self._iskclimatology.SetProperty('DoPiecewiseLinear', 0)
        else:
            self._iskclimatology.SetProperty('DoPiecewiseLinear', 1)

        self._update_internal_clim()

        self._climatology_created = True

    @wrap_skif_functionfail
    def _update_internal_clim(self):
        self._iskclimatology.SetProperty('Heights', self._altitudes)
        for species, values in self._values.items():
            if hasattr(skif, 'AddGeneratedGlobalClimatologyHandleIfNotExists'):
                skif.AddGeneratedGlobalClimatologyHandleIfNotExists(species)
            else:
                logging.warning('SASKTRANIF Version is out of date')
            self._iskclimatology.SetPropertyUserDefined(species, values)

    def __getitem__(self, item):
        return self._values[item]

    @wrap_skif_functionfail
    def __setitem__(self, item, value):
        self._values[item] = ArrayWithCallback(value)
        self._iskclimatology.SetPropertyUserDefined(item, value)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if hasattr(self, '_climatology_created') and self._climatology_created:
            self._update_internal_clim()


class ClimatologyUserDefined2D(Climatology):
    """
    A two dimensional user defined climatology in altitude and angle within a user specified plane.

    Parameters
    ----------
    angles_deg: np.ndarray
        One dimensional array of the angular grid in degrees
    alts_m: np.ndarray
        One dimensional array of the height grid in meters
    values: dict
        dictionary of species, value pairs. Values should be two dimensional arrays of shape (len(angles_deg), len(alts_m))
    reference_vector: np.ndarray
        Length 3 unit vector defining the location where the angle is 0, i.e., the x-axis of the plane
    normal_vector: np.ndarray
        Length 3 unit vector defining the normal vector of the plane.
    """
    def __init__(self, angles_deg: np.ndarray, alts_m: np.ndarray, values: dict, reference_vector: np.ndarray,
                 normal_vector: np.ndarray):
        self._angular_grid = angles_deg
        self._alt_grid = alts_m
        self._reference = reference_vector
        self._normal = normal_vector
        self._values = {key: ArrayWithCallback(val) for key, val in values.items()}

        super().__init__('USERDEFINED_PROFILE_PLANE')

        self._climatology_created = False
        self._create_climatology()

    def supported_species(self):
        return list(self._values.keys())

    @wrap_skif_functionfail
    def _create_climatology(self):
        self._iskclimatology.SetProperty('normalandreference', np.concatenate((self._reference, self._normal)))

        self._update_internal_clim()

        self._climatology_created = True

    @wrap_skif_functionfail
    def _update_internal_clim(self):
        self._iskclimatology.SetProperty('Heights', self._alt_grid)
        self._iskclimatology.SetProperty('Angles', self._angular_grid)
        for species, values in self._values.items():
            if hasattr(skif, 'AddGeneratedGlobalClimatologyHandleIfNotExists'):
                skif.AddGeneratedGlobalClimatologyHandleIfNotExists(species)
            else:
                logging.warning('SASKTRANIF Version is out of date')
            self._iskclimatology.SetPropertyUserDefined(species, values.flatten())

    def __getitem__(self, item):
        return self._values[item]

    @wrap_skif_functionfail
    def __setitem__(self, item, value):
        self._values[item] = ArrayWithCallback(value)
        self._iskclimatology.SetPropertyUserDefined(item, value)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if hasattr(self, '_climatology_created') and self._climatology_created:
            self._update_internal_clim()


class ClimatologyUserDefined3D(Climatology):
    """ A user defined climatology class for profiles which have a geographic dependence.
    
    For more information see `sk.ClimatologyUserDefined`.
    
    Parameters
    ----------
    lats : np.ndarray
        The latitude grid (1D) for the climatology. The latitudes for the region will be queried must be defined. 
        Latitudes don't need to be evenly spaced.
    lons : np.ndarray
        The longitude grid (1D) for the climatology. Must span 0 to 360 longitude (but doesn't need even spacing).
    alts : np.ndarray
        The altitude grid for the climatology in meters. Altitudes don't need to be evenly spaced. Altitudes at the 
        top of that atmosphere (TOA) and the ground (elevation) must be defined.
    values:
        A 3D array of values with dimensions: latitude, longitude, altitude.
    interp: str, optional
        One of 'linear' or 'log'. Defines the interpolation space. Default is 'linear'
    spline: bool, optional
        One of True of False.  If True then a quadratic spline will be used to interpolate, if False then piecewise
        linear interpolation is done.  Default is False.
    """
    def __init__(self, lats: np.ndarray, lons: np.ndarray, alts: np.ndarray, values: dict, interp: str='linear', spline: bool=False):
        super().__init__('USERDEFINED_PROFILE3D_LATLONHEIGHT')
        self._latitudes = lats
        self._longitudes = lons
        self._altitudes = alts

        self._values = {key: ArrayWithCallback(val) for key, val in values.items()}

        for val in self._values.values():
            val.add_modified_callback(self._update_internal_clim)

        self._interp = interp
        self._spline = spline

        self._climatology_created = False
        self._create_climatology()

    @wrap_skif_functionfail
    def _create_climatology(self):
        self._update_internal_clim()
        self._climatology_created = True

    @wrap_skif_functionfail
    def _update_internal_clim(self):
        self._iskclimatology.SetProperty('Heights', self._altitudes)
        self._iskclimatology.SetProperty('Latitudes', self._latitudes)
        self._iskclimatology.SetProperty('Longitudes', self._longitudes)
        for species, values in self._values.items():
                self._iskclimatology.SetPropertyUserDefined(species, values.flatten())

    def __getitem__(self, item):
        return self._values[item]

    @wrap_skif_functionfail
    def __setitem__(self, item, value):
        self._values[item] = ArrayWithCallback(value)
        self._iskclimatology.SetPropertyUserDefined(item, value)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if hasattr(self, '_climatology_created') and self._climatology_created:
            self._update_internal_clim()

    def supported_species(self):
        return [species for species in self._values.keys()]


class Labow(Climatology):
    """
    The Labow volume mixing ratio climatology of ozone. This is a climatology of the volume mixing ratio from
    0 km to 60 km in 1 km increments for 18 latitudes in steps of 10 degrees from -85 to +85 for each month of the year.
    The climatology seems to be unpublished as all references refer to an unpublished piece of work by McPeters and
    Labow in 2002 or 2003. We have extended the climatology above 60 km using the differences in scale height between
    the neutrals (7 km scale height ) and ozone (4.5 km scale height) to calculate a new scale height to extrapolate
    the signal at 60 km. We have also extended the volume mixing ratio so it can be converted to an ozone density.
    This requires an atmospheric number density climatology for which the MSIS90 climatology is used:
    """
    def __init__(self):
        super().__init__('O3LABOW')


class Pratmo(Climatology):
    """
    A climatology for the NO2 number density based upon the Prather photo-chemical box model. This climatological
    model is based upon monthly mean solutions to the box model for a range of latitudes, local solar times and
    altitudes.
    """
    def __init__(self):
        super().__init__('NO2PRATMO')

    def __repr__(self):
        supported_species = ['SKCLIMATOLOGY_NO2_CM3']

        representation = ('SasktranIF Climatology: {}\n'
                          'Supported Species: {}'.format(self._name, supported_species))

        return representation

    def supported_species(self):
        return ['SKCLIMATOLOGY_NO2_CM3']


class MSIS90(Climatology):
    """
    A climatology that implements the MSIS-90 atmospheric model. The current implementation is not complete in the
    sense that we do not attempt to enter the correct solar activity parameters to the model. Rather we use it as a
    quick, efficient and robust atmosphere. It always provides reasonable climatological estimates of pressure and
    temperature.
    """
    def __init__(self, include_o2=False):
        super().__init__('MSIS90')
        self._include_o2 = include_o2

    def supported_species(self):
        species = super().supported_species()

        if not self._include_o2:
            species.remove('SKCLIMATOLOGY_O2_CM3')

        return species


class ECMWF(Climatology):
    def __init__(self):
        super().__init__('ECMWF')

    def __repr__(self):
        representation = ('SasktranIF Climatology: {}\n'
                          'Supported Species: {}'.format(self._name, ['SKCLIMATOLOGY_PRESSURE_PA',
                                                                      'SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3',
                                                                      'SKCLIMATOLOGY_TEMPERATURE_K']))
        return representation

    @staticmethod
    def _validate_registry():
        registry_config = config.load_sasktran_registry()

        ecmwf_folder = registry_config['software']['usask-arg']['climatology']['ecmwf']['storage']['erainterimdir']

        if ecmwf_folder.lower() == 'undefined':
            user_config = config.user_config_file_location()
            raise IOError('ECMWF Folder is not set, please add ecmwf_directory: xxxx to the file {}'.format(user_config))


class GloSSAC(ClimatologyUserDefined):
    """
    The global stratospheric aerosol extinction climatology at 8 wavelengths from 386-3400nm that spans from 1979-2016
    and incorporates data from numerous satellite datasets. Interpolations in latitude and time use nearest neigbour.

    Note about wavelengths:
        386, 452, 525 and 1020 are the merged aerosol extinction datasets, although 386 and 452nm are not generally
        recommended. 525 and 1020nm channels provide the most robust extinction values so for radiative transfer
        calculations it is recommended to use one of these two channels.

        Other channels (750, 1257 and 3400) are from individual instruments.

    As the GloSSAC climatology is extinction based the species_name is 'SKCLIMATOLOGY_AEROSOL_EXTINCTIONPERKM', although
    the skif_object is converted to number density for use in sasktran.

    Parameters
    ----------
    wavelength_nm: float
        wavelength of the returned extinction. Possible values are: 386, 452, 525, 750, 1020, 1257, 3400
    extend: bool
        If true the extinction profile is extended above and below the valid range to avoid sharp distcontinuities.
        Profiles are assumed to exponenetially decay outside the valid range. Default is False where all non-valid
        regions are set to zero. The class variables decay_below and decay_above can be used to adjust the decay rates
        below and above the valid range respectively
    """
    def __init__(self, wavelength_nm: int=525, extend=False):

        from netCDF4 import Dataset
        from configparser import ConfigParser
        import os

        settings = ConfigParser()
        settings.read(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'config.ini'))
        self._filename = os.path.normpath(settings.get('Aerosol', 'GloSSACFile'))
        try:
            self.dataset = Dataset(self._filename, 'r')
        except OSError:
            # Try the global sasktran config file instead
            from sasktran.config import load_user_config, user_config_file_location

            try:
                self._filename = load_user_config()['glossac_file']
                self.dataset = Dataset(self._filename, 'r')
            except KeyError:
                raise IOError('error reading the GloSSAC file, ' + self._filename + ' you may need to update the '
                              'location of the GloSSAC climatology in {}'.format(user_config_file_location()))
            except OSError:
                raise IOError('error reading the GloSSAC file, ' + self._filename + ' you may need to update the '
                              'key glossac_file in {}'.format(user_config_file_location()))

        altitudes = np.arange(0, 100000, 500)
        self._species_name = 'SKCLIMATOLOGY_AEROSOL_EXTINCTIONPERKM'
        self.wavelength_nm = wavelength_nm
        self.extend = extend
        self.decay_below = 1e-4
        self.decay_above = 5e-4

        super().__init__(altitudes=altitudes, values={self._species_name: altitudes * 0}, interp='linear', spline=False)

    def get_parameter(self, species: str, latitude: float, longitude: float, altitudes: np.ndarray, mjd: float):

        if species not in self._values.keys():
            raise SasktranError('species not supported')

        # convert the time in the nc file to mjds
        year = np.floor(np.array(self.dataset['time'][:])).astype(int)
        month = np.round(np.array(self.dataset['time'][:] % 1 * 12)).astype(int) + 1
        time = np.array([(np.datetime64(str(y + 1979) + '-' + str(m).zfill(2) + '-15') - np.datetime64('1858-11-17')) /
                         np.timedelta64(1, 'D') for y, m in zip(year, month)])

        lat_idx = np.argmin(np.abs(latitude - self.dataset['lat'][:]))
        date_idx = np.argmin(np.abs(mjd - time))

        # in the glossac dataset 2 525nm channels exist (2 is the merged, 3 is OSIRIS only so use 2)
        if np.abs(self.wavelength_nm - 525) <= \
                np.min(np.abs(self.wavelength_nm - self.dataset['measurement_wavelengths'][:] * 1000)):
            wavel_idx = 2
        else:
            wavel_idx = np.argmin(np.abs(self.wavelength_nm - self.dataset['measurement_wavelengths'][:] * 1000))

        ext = self.dataset['Measurements_extinction'][:, lat_idx, date_idx, wavel_idx]
        if type(ext) is np.ma.core.MaskedArray:
            ext = ext.data
        good = (ext != self.dataset['Measurements_extinction'].getncattr('missing_value')) & ~np.isnan(ext)
        if sum(good) == 0:
            raise ValueError('Extinction profile at this wavelength is entirely nans')
        ext = np.interp(altitudes, self.dataset['altitude1'][good] * 1000, ext[good])
        if hasattr(altitudes, '__len__'):
            if self.extend:
                # decay the aerosol above the top point faster than the background
                msis = MSIS90()
                high_alts = altitudes > self.dataset['altitude1'][good][-1] * 1000
                air = msis.get_parameter('SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', latitude, longitude,
                                         altitudes[high_alts], mjd)
                air_top = msis.get_parameter('SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', latitude, longitude,
                                             self.dataset['altitude1'][good][-1] * 1000, mjd)
                last = np.where(~high_alts)[0][-1]
                ext[high_alts] = ext[high_alts][0] * air / air_top * \
                    np.exp(self.decay_above * -(altitudes[high_alts] - altitudes[last]))

                # exponentially decay the profile below the lower bound
                low_alts = altitudes < self.dataset['altitude1'][good][0] * 1000
                first = np.where(~low_alts)[0][0]
                ext[low_alts] = ext[first] * np.exp(self.decay_below * (altitudes[low_alts] - altitudes[first]))
            else:
                ext[altitudes > self.dataset['altitude1'][-1] * 1000] = 0.0
                ext[altitudes < self.dataset['altitude1'][0] * 1000] = 0.0

        else:
            if altitudes > self.dataset['altitude1'][-1] * 1000:
                ext = 0.0
            if altitudes < self.dataset['altitude1'][0] * 1000:
                ext = 0.0

        return ext

    def skif_object(self, **kwargs):

        clim_values = dict()

        engine = kwargs['engine']
        reference_point = engine.model_parameters['referencepoint']
        latitude = reference_point[0]
        longitude = reference_point[1]
        mjd = reference_point[3]

        opt_prop = kwargs['opt_prop']
        xsec = np.array([opt_prop.calculate_cross_sections(MSIS90(), latitude, longitude, a, mjd,
                                                           self.wavelength_nm).scattering for a in self._altitudes]).flatten()
        clim_values['SKCLIMATOLOGY_AEROSOL_CM3'] = self.get_parameter(self._species_name, latitude, longitude,
                                                                      self._altitudes, mjd) / xsec / 1e5
        userdef_clim = ClimatologyUserDefined(self._altitudes, clim_values)

        return userdef_clim.skif_object()