import sasktranif.sasktranif as skif
import numpy as np
import sasktran as sk
import logging
from sasktran.exceptions import wrap_skif_functionfail
from collections import namedtuple
from sasktran.util import DictWithCallback, LowercaseKeyDictWithCallback, to_iter


class Engine(object):
    """ 
    Base class for objects which can perform radative transfer calculations.
    """
    
    @wrap_skif_functionfail
    def __init__(self, name: str, geometry: sk.Geometry = None, atmosphere: sk.Atmosphere = None,
                 wavelengths: list = None, options: dict = None):
        """ Base class constructor. The base class cannot be instantiated.
        """
        self._name = name
        self._iskengine = skif.ISKEngine(self._name)
        self._wavelengths = wavelengths
        if geometry is not None:
            self._geometry = geometry
        else:
            self._geometry = sk.Geometry()

        if atmosphere is not None:
            self._atmosphere = atmosphere
        else:
            self._atmosphere = sk.Atmosphere()
        self._needs_reconfigure = True
        self._options = DictWithCallback(options) if options is not None else DictWithCallback()
        self._options.add_modified_callback(self._set_needs_reconfigure)

    def skif_object(self, **kwargs):
        """ Get the underlying ISK object.
        
        Returns
        -------
        skif.ISKEngine
            Underlying ISKEngine object
        """
        return self._iskengine

    @property
    def options(self):
        """
        dict : Dictionary map of additional engine options.

        Examples
        --------
        >>> import sasktran as sk
        >>> engine = sk.EngineHR()
        >>> engine.options['numordersofscatter'] = 1
        """
        return self._options

    @property
    def atmosphere(self):
        """
        sasktran.Atmosphere : The engines atmosphere definition object.
        """
        return self._atmosphere

    @atmosphere.setter
    def atmosphere(self, value):
        self._atmosphere = value
        self._needs_reconfigure = True

    @property
    def geometry(self):
        """
        sasktran.Geometry : The engines geometry definition object.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        self._geometry = value
        self._needs_reconfigure = True

    @property
    def wavelengths(self):
        """
        np.ndarray or float : Wavelengths in nm to perform the radiative transfer calculation at.

        Examples
        --------
        >>> import sasktran as sk
        >>> engine = sk.EngineHR()
        >>> engine.wavelengths = 600 # Single wavelength
        >>> engine.wavelengths = [600, 340] # Multiple wavelengths
        """
        return self._wavelengths

    @wavelengths.setter
    def wavelengths(self, value):
        self._wavelengths = value

    @wrap_skif_functionfail
    def calculate_radiance(self):
        """ Calculate radiances. 

        Returns
        -------
        np.ndarray **or** np.ndarray, np.ndarray
            The first np.ndarray is always the computed radiances. The shape of the computed 
            radiances is (W, L) where W is the number of wavelengths, and L is the number of 
            lines of sight. 
            If ``atmosphere.wf_species != None`` then a second np.ndarray will be returned 
            containing the computed weighting functions. The shape of this np.ndarray is
            engine specific but typically (W, L, A) where A is the number of altitudes at which
            at which the weighting functions were calculated.
        """
        if self._needs_reconfigure or self._atmosphere.has_changed or self._geometry.has_changed:
            self._iskengine = skif.ISKEngine(self._name)
            self._add_atmosphere_wf()
            self._add_lines_of_sight()
            self._add_options()
            self._initialize_model()
            self._add_atmosphere()
            self._needs_reconfigure = False
            self._geometry._added_to_engine()
            self._atmosphere._added_to_engine()

        if self._wavelengths is not None:
            self._iskengine.SetWavelengths(np.atleast_1d(self._wavelengths))
        else:
            raise ValueError("Property wavelengths needs to be set before calling Engine.calculate_radiance()")

        return self._iskengine.CalculateRadiance()[1]

    @wrap_skif_functionfail
    def _initialize_model(self):
        self._iskengine.InitializeModel()

    def _set_needs_reconfigure(self):
        self._needs_reconfigure = True

    @wrap_skif_functionfail
    def _add_lines_of_sight(self):
        for los in to_iter(self._geometry.lines_of_sight):
            self._iskengine.AddLineOfSight(los.mjd, los.observer, los.look_vector)

        if self._geometry.sun is not None:
            self.options['setsun'] = self._geometry.sun

    @wrap_skif_functionfail
    def _add_atmosphere(self):
        for _, species in self._atmosphere.species.items():
            self._iskengine.AddSpecies(species.species, species.climatology.skif_object(engine=self, opt_prop=species.optical_property),
                                       species.optical_property.skif_object(engine=self))
        if self._atmosphere.atmospheric_state is not None:
            self._iskengine.SetAtmosphericState(self._atmosphere.atmospheric_state.skif_object())
        if self._atmosphere.brdf is not None:
            self._iskengine.SetBRDF(self._atmosphere.brdf.skif_object(engine=self))

    @wrap_skif_functionfail
    def _add_atmosphere_wf(self):
        if self._atmosphere.wf_species is not None:
            if self._atmosphere.wf_species in self._atmosphere.species:
                op = self._atmosphere.species[self._atmosphere.wf_species].optical_property.skif_object()
                self._iskengine.SetProperty('wfspecies', op)
                if 'calcwf' not in [key.lower() for key in self._options]:
                    self._iskengine.SetProperty('calcwf', 2)
            else:
                raise ValueError("WFspecies '%s' not found in species list" % self._atmosphere.wf_species)

    @wrap_skif_functionfail
    def _add_options(self):
        for key, val in self._options.items():
            self._iskengine.SetProperty(key, val)

    @wrap_skif_functionfail
    def _reset_engine(self):
        self._iskengine = skif.ISKEngine(self._name)

        self._needs_reconfigure = True


class EngineHR(Engine):
    """
    Examples
    --------
    >>> import sasktran as sk
    >>> # configure your geometry
    >>> geometry = sk.VerticalImage()
    >>> geometry.from_sza_saa(60, 60, 0, 0, [10, 20, 30, 40], 54372, 0)
    >>> # configure your atmosphere
    >>> atmosphere = sk.Atmosphere()
    >>> atmosphere['rayleigh'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
    >>> atmosphere['o3'] = sk.Species(sk.O3DBM(), sk.Labow())
    >>> atmosphere['no2'] = sk.Species(sk.NO2Vandaele1998(), sk.Pratmo())
    >>> atmosphere.brdf = sk.Lambertian(0.3)
    >>> # make your engine
    >>> engine = sk.EngineHR(geometry=geometry, atmosphere=atmosphere, wavelengths=[350, 400, 450])
    >>> rad = engine.calculate_radiance()
    >>> print(rad) #doctest: +ELLIPSIS
    [[ 0.11674...  0.10872...  0.04904...  0.01456...]
     [ 0.10837...  0.08537...  0.02897...  0.00804...]
     [ 0.10318...  0.06396...  0.01829...  0.00483...]]
    """
    def __init__(self, geometry: sk.Geometry = None, atmosphere: sk.Atmosphere = None,
                 wavelengths: list = None, options: dict = None):
        super().__init__('hr', geometry, atmosphere, wavelengths, options)

        self._model_parameters = dict()

    def _initialize_model(self):
        super()._initialize_model()

        self._model_parameters['referencepoint'] = self._iskengine.GetProperty('ReferencePoint')[1]
        self._model_parameters['sun'] = self._iskengine.GetProperty('sun')[1]

    def _add_atmosphere(self):
        super()._add_atmosphere()

        # Add in any atmospheric emissions
        for species, emission in self._atmosphere.emissions.items():
            self._iskengine.AddEmission(species, emission.skif_object(atmosphere=self._atmosphere, engine=self))

    def _add_lines_of_sight(self):
        super()._add_lines_of_sight()

        if self._geometry.reference_point is not None:
            self._iskengine.SetProperty('setreferencepoint', self._geometry.reference_point)

    @property
    def model_parameters(self):
        return self._model_parameters

    @wrap_skif_functionfail
    def calculate_radiance(self):
        rad = super().calculate_radiance()
        if self._atmosphere.wf_species is not None:
            wf = self._iskengine.GetWeightingFunctions()[1]
            output = namedtuple('EngineOutput', ['radiance', 'weighting_function'])
            return output(radiance=rad, weighting_function=wf)
        else:
            return rad


class EngineDO(Engine):
    """
    Examples
    --------
    >>> import sasktran as sk
    >>> # Configure the geometry
    >>> geometry = sk.NadirGeometry()
    >>> tempo = sk.Geodetic()
    >>> tempo.from_lat_lon_alt(0, -100, 35786000)
    >>> lats = np.linspace(20, 50, 4)
    >>> lons = np.linspace(-110, -90, 3)
    >>> lats, lons = np.meshgrid(lats, lons, indexing='ij')
    >>> geometry.from_lat_lon(mjd=57906.63472, observer=tempo, lats=lats, lons=lons)
    >>> # Configure the atmosphere
    >>> atmosphere = sk.Atmosphere()
    >>> atmosphere['rayleigh'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
    >>> atmosphere['o3'] = sk.Species(sk.O3DBM(), sk.Labow())
    >>> atmosphere['no2'] = sk.Species(sk.NO2Vandaele1998(), sk.Pratmo())
    >>> atmosphere.brdf = sk.Kokhanovsky()
    >>> # Calcualte radiance
    >>> engine = sk.EngineDO(geometry=geometry, atmosphere=atmosphere, wavelengths=[350, 450])
    >>> rad = engine.calculate_radiance()
    >>>
    >>> # print radiances
    >>> print(rad[0].reshape((4, 3))) # radiance at 350nm
    [[ 0.15726175  0.19468353  0.22767423]
     [ 0.1666762   0.20027906  0.22962238]
     [ 0.17116222  0.1995542   0.22411801]
     [ 0.17099479  0.1932349   0.21235274]]
    >>> print(rad[1].reshape((4, 3))) # radiance at 450nm
    [[ 0.15884236  0.19815234  0.23287579]
     [ 0.16822132  0.20341264  0.23415734]
     [ 0.17226753  0.20184478  0.22736965]
     [ 0.1716201   0.19441591  0.21402527]]
    """
    def __init__(self, geometry: sk.Geometry = None, atmosphere: sk.Atmosphere = None,
                 wavelengths: list = None, options: dict = None):
        super().__init__('do', geometry, atmosphere, wavelengths, options)
        self._model_parameters = dict()
        self._wf_shape = None
        self._options = LowercaseKeyDictWithCallback(options) if options is not None else DictWithCallback()
        self._options.add_modified_callback(self._set_needs_reconfigure)

    def _initialize_model(self):
        super()._initialize_model()
        self._model_parameters['sun'] = self._iskengine.GetProperty('sun')[1]
        self.model_parameters['numstreams'] = int(self._iskengine.GetProperty('numstreams')[1])
        self.model_parameters['numlayers'] = int(self._iskengine.GetProperty('numlayers')[1])
        self.model_parameters['numphasefexpansionterms'] = \
            int(self._iskengine.GetProperty('numphasefexpansionterms')[1])
        self.model_parameters['numbrdfexpansionterms'] = int(self._iskengine.GetProperty('numbrdfexpansionterms')[1])

    @wrap_skif_functionfail
    def _add_atmosphere_wf(self):
        if self.atmosphere.wf_species is not None:
            wf_species_handles = []
            for a_species in np.atleast_1d(self.atmosphere.wf_species):
                if not (a_species in self.atmosphere.species):
                    err_msg = "Weighting function species '{}' was not in the species list".format(a_species)
                    raise ValueError(err_msg)
                wf_species_handles.append(self.atmosphere.species[a_species].species)
            self.options['wfclimatologies'] = " ".join(wf_species_handles)
            if not ('wfaltitudes' in self.options):
                self.options['wfaltitudes'] = np.linspace(500, 99500, 100)
            if not ('wfwidths' in self.options):
                self.options['wfwidths'] = np.array([1000] * 100)
            if hasattr(self.atmosphere.wf_species, '__iter__') and not isinstance(self.atmosphere.wf_species, str):
                self._wf_shape = (
                    len(self.wavelengths),
                    len(self.geometry.lines_of_sight),
                    len(self.atmosphere.wf_species),
                    len(self.options['wfaltitudes'])
                )
            else:
                self._wf_shape = (
                    len(self.wavelengths),
                    len(self.geometry.lines_of_sight),
                    len(self.options['wfaltitudes'])
                )

    @property
    def model_parameters(self):
        return self._model_parameters

    @wrap_skif_functionfail
    def _add_lines_of_sight(self):
        for los in to_iter(self._geometry.lines_of_sight):
            self._iskengine.AddLineOfSight(los.mjd, los.observer, los.look_vector)

        if self._geometry.sun is not None:
            self.options['sun'] = self._geometry.sun
        if self._geometry.reference_point is not None:
            logging.warning('the reference point set in the geometry object is ignored by EngineDO')

    @wrap_skif_functionfail
    def calculate_radiance(self):
        rad = super().calculate_radiance()
        if self._atmosphere.wf_species is not None:
            wf = self._iskengine.GetWeightingFunctions()[1].reshape(self._wf_shape)
            output = namedtuple('EngineOutput', ['radiance', 'weighting_function'])
            return output(radiance=rad, weighting_function=wf)
        else:
            return rad

    @wrap_skif_functionfail
    def get_reference_points(self):
        reference_points = np.ndarray((len(self._geometry.lines_of_sight), 4))
        for i, los in enumerate(self._geometry.lines_of_sight):
            ok, ref_pt = self._iskengine.GetProperty('referencepoint[{}]'.format(i))
            reference_points[i, :] = ref_pt
        return reference_points

    @property
    def num_streams(self):
        return self.model_parameters.get('numstreams')

    @num_streams.setter
    def num_streams(self, n: int):
        self.options['numstreams'] = n

    @property
    def num_layers(self):
        return self.model_parameters.get('numlayers')

    @num_layers.setter
    def num_layers(self, n: int):
        self.options['numlayers'] = n

    @property
    def num_phasef_quadrature_terms(self):
        return self.model_parameters.get('numphasefexpansionterms')

    @num_phasef_quadrature_terms.setter
    def num_phasef_quadrature_terms(self, n: int):
        self.options['numphasefexpansionterms'] = n

    @property
    def num_brdf_quadrature_terms(self):
        return self.model_parameters.get('numbrdfexpansionterms')

    @num_brdf_quadrature_terms.setter
    def num_brdf_quadrature_terms(self, n: int):
        self.options['numbrdfexpansionterms'] = n