import sasktran as sk
import numpy as np
import logging
from sasktran.util import ListWithCallback
from sasktran.geodetic import Geodetic


class Geometry(object):
    """
    Class which represents the Geometry for a radiative transfer calculation.  Currently this is made up of
    a set of :py:class:`sasktran.LineOfSight` which define the viewing geometry, and a :py:attr:`sasktran.Geometry.sun` 
    property which controls the solar position.  In the future this class may be extended to also control options 
    such as the top of the atmosphere, and the 'ground' altitude.

    Note that it is not always necessary to explicitly set the sun position.  If no sun position is given then it
    is automatically calculated from the average `mjd` given in the lines of sight.

    This class can be thought of as the geometry interface to the radiative transfer model.  For simple
    problems, we can construct this class and modify the lines of sight directly, as is done in the example below.
    For more complicated problems it may be desirable to create a new class that inherits from this class, as is
    done in the :class:`VerticalImage` class.

    Examples
    --------
    >>> import sasktran as sk
    >>> geometry = sk.Geometry()
    >>> los1 = sk.LineOfSight(mjd=54832.5, observer=[3.6760131547888e+005, 1.0099763136400e+006, -6.871601202127e+006],\
                              look_vector=[2.884568631765662e-001, 7.925287180643269e-001,  5.372996083468238e-001]),
    >>> los2 = sk.LineOfSight(mjd=54832.5, observer=[3.6928085406796e+005, 1.0145908079886e+006, -6.870844156040e+006],\
                              look_vector=[2.88456863176566e-001, 7.925287180643269e-001,  5.372996083468238e-001])
    >>> geometry.lines_of_sight = [los1, los2]
    >>> print(geometry)
    Geometry object containing 2 lines of sight
    >>> geometry.sun = [0, 0, 1]
    >>> print(geometry)
    Geometry object containing 2 lines of sight
    Sun position forced to [0, 0, 1]
    """
    def __init__(self):
        self._los = ListWithCallback()
        self._los.add_modified_callback(self._set_has_changed)
        self._has_changed = False
        self._sun = None
        self._ref_point = None

    def __repr__(self):
        ret = "Geometry object containing {} lines of sight".format(len(self._los))
        if self._sun is not None:
            ret += '\nSun position forced to {}'.format(self._sun)
        return ret

    @property
    def lines_of_sight(self):
        """
        list : A list of :py:class:`sasktran.LineOfSight` in which to calculate radiances.  
        
        Examples
        --------
        >>> import sasktran as sk
        >>> geometry = sk.Geometry()
        >>> arg_office = sk.Geodetic()  # line of sight terminator
        >>> arg_office.from_lat_lon_alt(52.131638, -106.633873, 0)
        >>> observer = sk.Geodetic()   # line of sight observer
        >>> observer.from_lat_lon_alt(0, -100, 35786000)
        >>> look = (arg_office.location - observer.location)
        >>> look = look / np.linalg.norm(look)  # don't forget to normalize the look vector
        >>> # in this example we'll only add one line of sight (done below) but any number could be added to this list
        >>> geometry.lines_of_sight = [ \
             sk.LineOfSight(look_vector=look, observer=observer.location, mjd=57906), \
        ]
        """
        return self._los

    @lines_of_sight.setter
    def lines_of_sight(self, value):
        self._los = ListWithCallback(value)
        self._has_changed = True

    @property
    def sun(self):
        """
        np.ndarray((3,)) : Unit vector pointing to the sun in geodetic coordinates. See 
        :py:attr:`sasktran.Geodetic.location`.
        
        Examples
        --------
        >>> import sasktran as sk
        >>> geometry = sk.Geometry()
        >>> sun = sk.Geodetic()
        >>> sun.from_lat_lon_alt(22.26666667, -48.98333333, 0) # altitude (argument 3) doesnt matter here
        >>> geometry.sun = sun.local_up # a geocentric unit vector pointing at the sun
        """
        return self._sun

    @sun.setter
    def sun(self, value):
        self._sun = value
        self._has_changed = True

    @property
    def reference_point(self):
        """
        np.ndarray((3,)) : Geodetic coordinate of where (geographic location) the engine should poll 
        climatology and surface properties. See :py:attr:`sasktran.Geodetic.location`.
        
        Notes
        -----
        Reference points semantics are engine specific. Refer to your engines specific documentation for details.
        """
        return self._ref_point

    @reference_point.setter
    def reference_point(self, value):
        self._ref_point = value
        self._has_changed = True

    @property
    def has_changed(self):
        return self._has_changed

    def _set_has_changed(self):
        self._has_changed = True

    def _added_to_engine(self):
        self._has_changed = False


class VerticalImage(Geometry):
    """
    A specialized :class:`Geometry` class which has various convenience methods to construct lines of sight
    corresponding to a vertical image of the atmosphere based on input solar zenith angles.

    Examples
    --------
    >>> from sasktran.geometry import VerticalImage
    >>> geometry = VerticalImage()
    >>> geometry.from_sza_saa(60, 60, 0, 0, [10, 20, 30, 40], 54372, 0)
    >>> print(geometry)
    Geometry object containing 4 lines of sight
    Sun position forced to [ 0.5        0.75       0.4330127]
    """
    def __init__(self):
        super().__init__()

        self.reftp = None
        self.reflook = None
        self.refmjd = None

    def from_sza_saa(self, sza, saa, lat, lon, tanalts_km, mjd, locallook, satalt_km=600, refalt_km=20):
        """
        Constructs the geometry from a given solar zenith angle and solar azimuth angle

        Parameters
        ----------
        sza : float
            Solar zenith angle in degrees of the tangent point at refalt_km.
        saa : float
            Solar azimuth angle in degrees of the tangent point at refalt_km.
        lat : float
            Latitude in degrees (-90 to 90)
        lon : float
            Longitude in degrees
        tanalts_km : np.ndarray
            Tangent altitudes in km
        mjd : float
            Modified Julian Date for the measurements
        locallook : float
            Angle in degrees defining the look direction on the surface of the Earth.  0 corresponds to true north,
            90 to East and so on.
        satalt_km : float, optional
            Altitude of the observers in km.  Default is 600 km.
        refalt_km : float, optional
            Altitude that the given solar zenith angle and solar azimuth angle are valid at.  Since the geometry
            represents a vertical image, the solar angles change for every line of sight.

        """
        geo = sk.Geodetic()

        geo.from_lat_lon_alt(lat, lon, refalt_km * 1000)

        self.reftp = geo.location
        west = geo.local_west
        south = geo.local_south
        up = geo.local_up

        sunh = -np.cos(np.deg2rad(saa)) * south - np.sin(np.deg2rad(saa)) * west
        self._sun = np.cos(np.deg2rad(sza)) * up + np.sin(np.deg2rad(sza)) * sunh

        self.reflook = -np.cos(np.deg2rad(locallook)) * south - np.sin(np.deg2rad(locallook)) * west

        a = (np.linalg.norm(self.reftp) + satalt_km * 1000) ** 2

        s = -np.dot(self.reftp, -self.reflook) + np.sqrt(
            np.dot(self.reftp, -self.reflook) ** 2 - np.linalg.norm(self.reftp) ** 2 + a)

        obs = self.reftp - s * self.reflook
        self.refobs = self.reftp - s * self.reflook
        self._los = []
        for alt in tanalts_km:
            look = geo.from_tangent_altitude(alt * 1000.0, obs, self.reflook)

            self._los.append(sk.LineOfSight(mjd, obs, look))
        self.refmjd = mjd
        self._calc_real_parameters()

    def from_sza_ssa(self, sza, ssa, lat, lon, tanalts_km, mjd, locallook, satalt_km=600, refalt_km=20):
        sza_r = np.deg2rad(sza)
        ssa_r = np.deg2rad(ssa)

        if np.cos(ssa_r) / np.sin(sza_r) > 1:
            logging.warning('Unphysical geometry, forcing scattering angle to be the maximum allowed, %f',
                            np.rad2deg(np.arccos(np.sin(sza_r))))
            saa = locallook
        else:
            saa = np.rad2deg(np.arccos(np.cos(ssa_r) / np.sin(sza_r))) + locallook

        self.from_sza_saa(sza, saa, lat, lon, tanalts_km, mjd, locallook, satalt_km, refalt_km)

    def _calc_real_parameters(self):
        geo = sk.Geodetic()
        geo.from_tangent_point(self.refobs, self.reflook)
        self.reftp = geo.location

        self.lon = geo.longitude
        self.lat = geo.latitude

        # SSA is sun dot look
        self.ssa = np.rad2deg(np.arccos(np.dot(self._sun, self.reflook)))
        # SZA is sun dot tp
        self.sza = np.rad2deg(np.arccos(np.dot(self._sun, self.reftp / np.linalg.norm(self.reftp))))
        # TODO: calc saa
        self.saa = None


class NadirGeometry(Geometry):
    def __init__(self):
        super().__init__()
        self._int_lat = None
        self._int_lon = None
        self._oza = None
        self._oaa = None
        self._sza = None
        self._saa = None

    def from_lat_lon(self, mjd, observer: Geodetic, lats, lons, elevations=0):
        """
        Construct lines of sight from arrays of points (latitude, longitude[, elevation]).
        
        Notes
        -----
        Arguments for this function are associated sequences (ie. the first observation is at (lats[0], lons[0], the
        second is at (lats[1], lons[1]) and so on). This means that if you have a geoegraphic grid of sample points, 
        these arrays should be flattened before being passed to this function.
        
        Arguments are broadcast together using numpy's broadcast rules.

        Examples
        --------
        >>> geometry = sk.NadirGeometry()
        >>> # make the look vector from TEMPO to the ARG office
        >>> tempo = sk.Geodetic()
        >>> tempo.from_lat_lon_alt(0, -100, 35786000)
        >>> geometry.from_lat_lon(mjd=57906, observer=tempo, lats=52.131638, lons=-106.633873, elevations=0)
        
        Parameters
        ----------
        mjd : float, np.ndarray
            Modified Julian Data of observations. If ``mjd`` is a ``float`` then this mjd will be used for each line of 
            sight. Else mjd's are associated element-wise.
        observer : sasktran.Geodetic or array of sasktran.Geodetic
            Observer locations.
        lats : float, np.ndarray
            Latitudes of observations. If this argument is a float then this latitude will be used for each line of 
            sight, otherwise latitudes will be flattened and associated element wise.
        lons : float, np.ndarray
            Same as ``lats`` but for longitudes.   
        elevations : float, np.ndarray
            Used to specify the ground elevation (in meters) for each sample point.
        """
        # broadcast arguments to sequence of sample points
        mjd = np.atleast_1d(mjd).astype(float)
        obs = np.atleast_1d(observer)
        lats = np.atleast_1d(lats).astype(float)
        lons = np.atleast_1d(lons).astype(float)
        elev = np.atleast_1d(elevations).astype(float)

        shape = np.broadcast(mjd, obs, lats, lons, elev).shape
        mjd = np.broadcast_to(mjd, shape).flatten()
        obs = np.broadcast_to(obs, shape).flatten()
        lats = np.broadcast_to(lats, shape).flatten()
        lons = np.broadcast_to(lons, shape).flatten()
        elev = np.broadcast_to(elev, shape).flatten()

        # calculate lines of sight
        self.lines_of_sight = []
        for latitude, longitude, observer, mj_date, elevation in zip(lats, lons, obs, mjd, elev):
            terminator = sk.Geodetic()
            terminator.from_lat_lon_alt(latitude, longitude, elevation)

            look = terminator.location - observer.location
            look = look / np.linalg.norm(look)

            self.lines_of_sight.append(
                sk.LineOfSight(mjd=mj_date, observer=observer.location, look_vector=look)
            )

    def from_fov(self, mjd: float, observer: Geodetic,
                 center: Geodetic, ntheta: int=1, nphi: int=1,
                 dtheta: float=0.0, dphi: float=0.0,
                 bearing: float=None):
        """
        Adds lines of sight by creating a rectangular grid of equally
        spaced angles surrounding a center line of sight. The center line of
        sight is  defined by the observer position and the center position.

        Parameters
        ----------
        observer : sk.Geodetic
            Position of the observer.
        center : sk.Geodetic
            The center of the bundle of lines of sight.
        ntheta : int
            The number of rows in the angle grid.
        nphi : int
            The number of columns in the angle grid.
        dtheta: float
            The spacing between columns in the angle grid.
        dphi : float
            The spacing between rows in the angle grid.
        bearing : float
            The bearing (degrees, CW from north) of the center column's
            intersection with the local horizontal at center. Defaults to None,
            which places the column in the plane containing the observer and
            the local vertical at center.
        """

        # get the center look vector
        look_center = center.location - observer.location
        look_center = look_center / np.linalg.norm(look_center)

        # get the axis perpendicular to the center column plane
        if bearing is None:
            col_axis = np.cross(look_center, center.local_up)
        else:
            cosb = np.cos(np.deg2rad(bearing))
            sinb = np.sin(np.deg2rad(bearing))
            local_horizontal = -(cosb * center.local_south + sinb * center.local_west)
            col_axis = np.cross(look_center, local_horizontal)

        # get the LOSs of the center column
        theta_range = (ntheta - 1) * dtheta
        theta = np.linspace(-theta_range / 2, theta_range / 2, ntheta)
        col = np.zeros((ntheta, 3))
        for i in range(ntheta):
            R = self._rotation_matrix(col_axis, theta[i])
            col[i, :] = np.dot(R, look_center)

        # fan the center column out into a grid
        phi_range = (nphi - 1) * dphi
        phi = np.linspace(-phi_range / 2, phi_range / 2, nphi)
        row_axis = np.cross(look_center, col_axis)
        looks = np.zeros((ntheta, nphi, 3))
        for j in range(nphi):
            R = self._rotation_matrix(row_axis, phi[j])
            looks[:, j, :] = np.dot(col, R)

        self._los = []
        for i in range(ntheta):
            for j in range(nphi):
                self._los.append(sk.LineOfSight(mjd, observer.location, looks[i, j, :]))

        self._calc_intersection()
        self._calc_los_angles()
        self._calc_sun_angles()
        self._has_changed = True

    def _calc_intersection(self):
        """
        Calculate the intersection of every line of sight with the earth.
        Fills with nans where no intersection occurs.
        """
        self._int_lon = []
        self._int_lat = []
        for los in self._los:
            geo = sk.Geodetic()
            i1, i2 = geo.altitude_intercepts(0.0, los.observer, los.look_vector)
            if i1 is None:
                self._int_lon.append(np.nan)
                self._int_lat.append(np.nan)
            else:
                geo.from_xyz(i1)
                self._int_lon.append(geo.longitude)
                self._int_lat.append(geo.latitude)

    def _calc_los_angles(self):
        """
        Calculates the zenith and azimuth angles of the observer from the
        ground intersection of every line of sight with a ground intersection
        """
        self._oza = []
        self._oaa = []

        for los, lon, lat in zip(self._los, self._int_lon, self._int_lat):
            if any(np.isnan([lat, lon])):
                self._oza.append(np.nan)
                self._oaa.append(np.nan)
            else:
                geo = sk.Geodetic()
                geo.from_lat_lon_alt(lat, lon, 0.0)
                self._oza.append(self._zenith(-los.look_vector, geo.local_up))
                self._oaa.append(self._azimuth(-los.look_vector, geo.local_south, geo.local_west))

    def _calc_sun_angles(self):
        """
        Calculates the zenith and azimuth angles of the sun from the ground
        intersection of every line of sight with a ground intersection.
        :return:
        """
        if self._sun is not None:
            self._sza = []
            self._saa = []

            for los, lon, lat in zip(self._los, self._int_lon, self._int_lat):
                if any(np.isnan([lat, lon])):
                    self._sza.append(np.nan)
                    self._saa.append(np.nan)
                else:
                    geo = sk.Geodetic()
                    geo.from_lat_lon_alt(lat, lon, 0.0)
                    self._sza.append(self._zenith(-self._sun, geo.local_up))
                    self._saa.append(self._azimuth(-self._sun, geo.local_south, geo.local_west))

    def _zenith(self, vec, up):
        dot = np.max([-1.0, np.min([1.0, np.dot(vec, up)])])
        return np.rad2deg(np.arccos(dot))

    def _azimuth(self, vec, south, west):
        return np.rad2deg(np.arctan2(np.dot(vec, -west), np.dot(vec, -south)))

    def _rotation_matrix(self, axis, angle):
        """
        Create a 3D rotation matrix using the Euler-Rodrigues formula

        :param axis: axis of rotation
        :param angle: angle to rotate around the axis (radians)
        :return: 3x3 rotation matrix
        """

        unit_axis = axis / np.linalg.norm(axis)

        half_angle_rad = angle / 2.0
        sin_half_angle_rad = np.sin(half_angle_rad)
        cos_half_angle_rad = np.cos(half_angle_rad)
        a = cos_half_angle_rad
        b = unit_axis[0] * sin_half_angle_rad
        c = unit_axis[1] * sin_half_angle_rad
        d = unit_axis[2] * sin_half_angle_rad

        R = np.zeros((3, 3))
        R[0, 0] = a * a + b * b - c * c - d * d
        R[0, 1] = 2 * (b * c - a * d)
        R[0, 2] = 2 * (b * d + a * c)
        R[1, 0] = 2 * (b * c + a * d)
        R[1, 1] = a * a + c * c - b * b - d * d
        R[1, 2] = 2 * (c * d - a * b)
        R[2, 0] = 2 * (b * d - a * c)
        R[2, 1] = 2 * (c * d + a * b)
        R[2, 2] = a * a + d * d - b * b - c * c

        return R

    @property
    def sun(self):
        return self._sun

    @sun.setter
    def sun(self, value):
        self._sun = value
        self._calc_sun_angles()
        self._has_changed = True

    @property
    def lines_of_sight(self):
        return self._los

    @lines_of_sight.setter
    def lines_of_sight(self, value):
        self._los = ListWithCallback(value)
        self._calc_intersection()
        self._calc_los_angles()
        self._calc_sun_angles()
        self._has_changed = True

    @property
    def latitude(self):
        return self._int_lat

    @property
    def longitude(self):
        return self._int_lon

    @property
    def solar_zenith(self):
        return self._sza

    @property
    def solar_azimuth(self):
        return self._saa

    @property
    def observer_zenith(self):
        return self._oza

    @property
    def observer_azimuth(self):
        return self._oaa